#include <cuda_fp16.h>
#include <cuda_fp16.hpp>
#include "activation.cuh"
#include <c10/cuda/CUDAGuard.h>
#include <ATen/cuda/CUDAContext.h>
#include "util.h"
#include "util.cuh"
#include "compat.cuh"
#include <cmath>

//__global__ __launch_bounds__(MAX_HEAD_DIM)
//void causal_conv1d_kernel
//(
//
//)
//{
//
//}
//

void causal_conv1d
(






)
{


}

